from typing import Any

argument_types = ...  # type: Any
converters = ...  # type: Any

def run(namespace=None, action_prefix='', args=None): ...
def fail(message, code=-1): ...
def find_actions(namespace, action_prefix): ...
def print_usage(actions): ...
def analyse_action(func): ...
def make_shell(init_func=None, banner=None, use_ipython=True): ...
def make_runserver(app_factory, hostname='', port=5000, use_reloader=False, use_debugger=False, use_evalex=True, threaded=False, processes=1, static_files=None, extra_files=None, ssl_context=None): ...
