from typing import Any

Int32Sizer = ...  # type: Any
UInt32Sizer = ...  # type: Any
SInt32Sizer = ...  # type: Any
Fixed32Sizer = ...  # type: Any
Fixed64Sizer = ...  # type: Any
BoolSizer = ...  # type: Any

def StringSizer(field_number, is_repeated, is_packed): ...
def BytesSizer(field_number, is_repeated, is_packed): ...
def GroupSizer(field_number, is_repeated, is_packed): ...
def MessageSizer(field_number, is_repeated, is_packed): ...
def MessageSetItemSizer(field_number): ...
def MapSizer(field_descriptor): ...
def TagBytes(field_number, wire_type): ...

Int32Encoder = ...  # type: Any
UInt32Encoder = ...  # type: Any
SInt32Encoder = ...  # type: Any
Fixed32Encoder = ...  # type: Any
Fixed64Encoder = ...  # type: Any
SFixed32Encoder = ...  # type: Any
SFixed64Encoder = ...  # type: Any
FloatEncoder = ...  # type: Any
DoubleEncoder = ...  # type: Any

def BoolEncoder(field_number, is_repeated, is_packed): ...
def StringEncoder(field_number, is_repeated, is_packed): ...
def BytesEncoder(field_number, is_repeated, is_packed): ...
def GroupEncoder(field_number, is_repeated, is_packed): ...
def MessageEncoder(field_number, is_repeated, is_packed): ...
def MessageSetItemEncoder(field_number): ...
def MapEncoder(field_descriptor): ...
