# Stubs for platform (Python 2)
#
# Based on stub generated by stubgen.

from typing import Any

__copyright__ = ...  # type: Any
DEV_NULL = ...  # type: Any

def libc_ver(executable=..., lib=..., version=..., chunksize=2048): ...
def linux_distribution(distname=..., version=..., id=..., supported_dists=..., full_distribution_name=1): ...
def dist(distname=..., version=..., id=..., supported_dists=...): ...

class _popen:
    tmpfile = ...  # type: Any
    pipe = ...  # type: Any
    bufsize = ...  # type: Any
    mode = ...  # type: Any
    def __init__(self, cmd, mode=..., bufsize=None): ...
    def read(self): ...
    def readlines(self): ...
    def close(self, remove=..., error=...): ...
    __del__ = ...  # type: Any

def popen(cmd, mode=..., bufsize=None): ...
def win32_ver(release=..., version=..., csd=..., ptype=...): ...
def mac_ver(release=..., versioninfo=..., machine=...): ...
def java_ver(release=..., vendor=..., vminfo=..., osinfo=...): ...
def system_alias(system, release, version): ...
def architecture(executable=..., bits=..., linkage=...): ...
def uname(): ...
def system(): ...
def node(): ...
def release(): ...
def version(): ...
def machine(): ...
def processor(): ...
def python_implementation(): ...
def python_version(): ...
def python_version_tuple(): ...
def python_branch(): ...
def python_revision(): ...
def python_build(): ...
def python_compiler(): ...
def platform(aliased=0, terse=0): ...
